// Copyright 2004, 2005 The Apache Software Foundation
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package hivemind.test.services;

import hivemind.test.services.impl.StringHolderImpl;

import org.apache.commons.logging.Log;
import org.apache.hivemind.ApplicationRuntimeException;
import org.apache.hivemind.ClassResolver;
import org.apache.hivemind.ErrorHandler;
import org.apache.hivemind.ErrorLog;
import org.apache.hivemind.Messages;
import org.apache.hivemind.Registry;
import org.apache.hivemind.ServiceImplementationFactoryParameters;
import org.apache.hivemind.impl.DefaultClassResolver;
import org.apache.hivemind.internal.Module;
import org.apache.hivemind.service.impl.BuilderClassResolverFacet;
import org.apache.hivemind.service.impl.BuilderErrorHandlerFacet;
import org.apache.hivemind.service.impl.BuilderErrorLogFacet;
import org.apache.hivemind.service.impl.BuilderFacet;
import org.apache.hivemind.service.impl.BuilderFactoryLogic;
import org.apache.hivemind.service.impl.BuilderLogFacet;
import org.apache.hivemind.service.impl.BuilderMessagesFacet;
import org.apache.hivemind.service.impl.BuilderParameter;
import org.apache.hivemind.service.impl.BuilderServiceIdFacet;
import org.apache.hivemind.test.AggregateArgumentsMatcher;
import org.apache.hivemind.test.ArgumentMatcher;
import org.apache.hivemind.test.HiveMindTestCase;
import org.apache.hivemind.test.TypeMatcher;
import org.easymock.MockControl;

/**
 * Tests for the standard {@link org.apache.hivemind.service.impl.BuilderFactory}service and
 * various implementations of {@link org.apache.hivemind.service.impl.BuilderFacet}.
 * 
 * @author Howard Lewis Ship
 */
public class TestBuilderFactory extends HiveMindTestCase
{
    private Object execute(ServiceImplementationFactoryParameters fp, BuilderParameter p)
    {
        return new BuilderFactoryLogic(fp, p).createService();
    }

    public void testSmartFacet() throws Exception
    {
        Registry r = buildFrameworkRegistry("SmartFacet.xml");

        SimpleService s = (SimpleService) r.getService(
                "hivemind.test.services.Simple",
                SimpleService.class);

        assertEquals(99, s.add(1, 1));
    }

    public void testInitializeMethodFailure() throws Exception
    {
        Registry r = buildFrameworkRegistry("InitializeMethodFailure.xml");

        Runnable s = (Runnable) r.getService("hivemind.test.services.Runnable", Runnable.class);

        interceptLogging("hivemind.test.services.Runnable");

        s.run();

        assertLoggedMessagePattern("Error at .*?: Unable to initialize service hivemind\\.test\\.services\\.Runnable "
                + "\\(by invoking method doesNotExist on "
                + "hivemind\\.test\\.services\\.impl\\.MockRunnable\\):");
    }

    public void testBuilderErrorHandlerFacet()
    {
        MockControl c = newControl(Module.class);
        Module m = (Module) c.getMock();

        ErrorHandler eh = (ErrorHandler) newMock(ErrorHandler.class);

        MockControl pc = newControl(ServiceImplementationFactoryParameters.class);
        ServiceImplementationFactoryParameters p = (ServiceImplementationFactoryParameters) pc
                .getMock();

        p.getInvokingModule();
        pc.setReturnValue(m);

        m.getErrorHandler();
        c.setReturnValue(eh);

        replayControls();

        BuilderFacet f = new BuilderErrorHandlerFacet();

        Object actual = f.getFacetValue(p, null);

        assertSame(eh, actual);

        verifyControls();
    }

    public void testSetErrorHandler() throws Exception
    {
        Registry r = buildFrameworkRegistry("SetErrorHandler.xml");

        ErrorHandlerHolder h = (ErrorHandlerHolder) r.getService(
                "hivemind.test.services.SetErrorHandler",
                ErrorHandlerHolder.class);

        assertNotNull(h.getErrorHandler());
    }

    public void testConstructErrorHandler() throws Exception
    {
        Registry r = buildFrameworkRegistry("ConstructErrorHandler.xml");

        ErrorHandlerHolder h = (ErrorHandlerHolder) r.getService(
                "hivemind.test.services.ConstructErrorHandler",
                ErrorHandlerHolder.class);

        assertNotNull(h.getErrorHandler());
    }

    public void testBuilderClassResolverFacet()
    {
        ClassResolver cr = (ClassResolver) newMock(ClassResolver.class);

        MockControl pc = newControl(ServiceImplementationFactoryParameters.class);
        ServiceImplementationFactoryParameters p = (ServiceImplementationFactoryParameters) pc
                .getMock();

        MockControl control = newControl(Module.class);
        Module module = (Module) control.getMock();

        p.getInvokingModule();
        pc.setReturnValue(module);

        module.getClassResolver();
        control.setReturnValue(cr);

        replayControls();

        BuilderClassResolverFacet fc = new BuilderClassResolverFacet();

        Object result = fc.getFacetValue(p, null);

        assertSame(cr, result);

        verifyControls();
    }

    public void testSetClassResolver() throws Exception
    {
        Registry r = buildFrameworkRegistry("SetClassResolver.xml");

        ClassResolverHolder h = (ClassResolverHolder) r.getService(
                "hivemind.test.services.SetClassResolver",
                ClassResolverHolder.class);

        assertNotNull(h.getClassResolver());
    }

    public void testConstructClassResolver() throws Exception
    {
        Registry r = buildFrameworkRegistry("ConstructClassResolver.xml");

        ClassResolverHolder h = (ClassResolverHolder) r.getService(
                "hivemind.test.services.ConstructClassResolver",
                ClassResolverHolder.class);

        assertNotNull(h.getClassResolver());
    }

    public void testAutowire()
    {
        MockControl fpc = newControl(ServiceImplementationFactoryParameters.class);
        ServiceImplementationFactoryParameters fp = (ServiceImplementationFactoryParameters) fpc
                .getMock();

        MockControl c = newControl(Module.class);
        Module module = (Module) c.getMock();

        ErrorHandler eh = (ErrorHandler) newMock(ErrorHandler.class);
        ClassResolver cr = new DefaultClassResolver();

        MockControl logc = newControl(Log.class);
        Log log = (Log) logc.getMock();

        MockControl messagesControl = newControl(Messages.class);
        Messages messages = (Messages) messagesControl.getMock();

        ErrorLog errorLog = (ErrorLog) newMock(ErrorLog.class);

        // Normally I try and get all the invocations into chronological
        // order ... but with this refactoring, that's painful; these
        // are in an order that appeases junit.

        fp.getLog();
        fpc.setReturnValue(log);

        fp.getServiceId();
        fpc.setReturnValue("foo.bar.Baz");

        fp.getInvokingModule();
        fpc.setReturnValue(module);

        module.getClassResolver();
        c.setReturnValue(cr);

        fp.getLog();
        fpc.setReturnValue(log);

        debug(fpc, fp, logc, log, "Autowired property log to " + log);

        fp.getInvokingModule();
        fpc.setReturnValue(module);

        module.getClassResolver();
        c.setReturnValue(cr);

        debug(fpc, fp, logc, log, "Autowired property classResolver to " + cr);

        fp.getInvokingModule();
        fpc.setReturnValue(module);

        module.getMessages();
        c.setReturnValue(messages);

        debug(fpc, fp, logc, log, "Autowired property messages to " + messages);

        fp.getInvokingModule();
        fpc.setReturnValue(module);

        module.getErrorHandler();
        c.setReturnValue(eh);

        debug(fpc, fp, logc, log, "Autowired property errorHandler to " + eh);

        fp.getServiceId();
        fpc.setReturnValue("foo.bar.Baz");

        debug(fpc, fp, logc, log, "Autowired property serviceId to foo.bar.Baz");

        fp.getErrorLog();
        fpc.setReturnValue(errorLog);

        debug(fpc, fp, logc, log, "Autowired property errorLog to " + errorLog);

        replayControls();

        BuilderParameter p = new BuilderParameter();

        p.setClassName(AutowireTarget.class.getName());
        p.addProperty(new BuilderLogFacet());
        p.addProperty(new BuilderClassResolverFacet());
        p.addProperty(new BuilderMessagesFacet());
        p.addProperty(new BuilderErrorHandlerFacet());
        p.addProperty(new BuilderServiceIdFacet());
        p.addProperty(new BuilderErrorLogFacet());

        AutowireTarget t = (AutowireTarget) execute(fp, p);

        assertSame(eh, t.getErrorHandler());
        assertSame(cr, t.getClassResolver());
        assertSame(messages, t.getMessages());
        assertSame(log, t.getLog());
        assertEquals("foo.bar.Baz", t.getServiceId());
        assertSame(errorLog, t.getErrorLog());

        verifyControls();
    }

    private void debug(MockControl fpc, ServiceImplementationFactoryParameters fp,
            MockControl logc, Log log, String string)
    {
        fp.getLog();
        fpc.setReturnValue(log);

        log.isDebugEnabled();
        logc.setReturnValue(true);

        log.debug(string);
    }

    /**
     * Test that BuilderFactory will invoke the "initializeService" method by default.
     */
    public void testAutowireInitializer()
    {
        MockControl c = newControl(Module.class);
        Module module = (Module) c.getMock();
        ClassResolver resolver = new DefaultClassResolver();

        MockControl fpc = newControl(ServiceImplementationFactoryParameters.class);
        ServiceImplementationFactoryParameters fp = (ServiceImplementationFactoryParameters) fpc
                .getMock();

        Log log = (Log) newMock(Log.class);

        fp.getLog();
        fpc.setReturnValue(log);

        fp.getServiceId();
        fpc.setReturnValue("foo");

        fp.getInvokingModule();
        fpc.setReturnValue(module);

        module.getClassResolver();
        c.setReturnValue(resolver);

        replayControls();

        BuilderParameter p = new BuilderParameter();

        p.setClassName(InitializeFixture.class.getName());

        InitializeFixture f = (InitializeFixture) execute(fp, p);

        // Check which method was actually invoked (if any)

        assertEquals("initializeService", f.getMethod());

        verifyControls();
    }

    /**
     * Test that BuilderFactory will invoke the named initializer.
     */
    public void testInitializer()
    {
        MockControl fpc = newControl(ServiceImplementationFactoryParameters.class);
        ServiceImplementationFactoryParameters fp = (ServiceImplementationFactoryParameters) fpc
                .getMock();

        MockControl c = newControl(Module.class);
        Module module = (Module) c.getMock();

        Log log = (Log) newMock(Log.class);

        fp.getLog();
        fpc.setReturnValue(log);

        fp.getServiceId();
        fpc.setReturnValue("foo");

        fp.getInvokingModule();
        fpc.setReturnValue(module);

        module.getClassResolver();
        c.setReturnValue(new DefaultClassResolver());

        replayControls();

        BuilderParameter p = new BuilderParameter();

        p.setClassName(InitializeFixture.class.getName());
        p.setInitializeMethod("initializeCustom");

        InitializeFixture f = (InitializeFixture) execute(fp, p);

        assertEquals("initializeCustom", f.getMethod());

        verifyControls();
    }

    public void testAutowireServices()
    {
        MockControl fpc = newControl(ServiceImplementationFactoryParameters.class);
        ServiceImplementationFactoryParameters fp = (ServiceImplementationFactoryParameters) fpc
                .getMock();

        MockControl mc = newControl(Module.class);
        Module module = (Module) mc.getMock();

        MockControl lc = newControl(Log.class);
        Log log = (Log) lc.getMock();

        fp.getLog();
        fpc.setReturnValue(log);

        fp.getServiceId();
        fpc.setReturnValue("foo");

        fp.getInvokingModule();
        fpc.setReturnValue(module);

        module.getClassResolver();
        mc.setReturnValue(new DefaultClassResolver());

        StringHolder h = new StringHolderImpl();

        module.getService(StringHolder.class);
        mc.setReturnValue(h);

        log.isDebugEnabled();
        lc.setReturnValue(false);

        replayControls();

        BuilderParameter parameter = new BuilderParameter();

        parameter.setClassName(ServiceAutowireTarget.class.getName());
        parameter.setAutowireServices(true);

        ServiceAutowireTarget service = (ServiceAutowireTarget) execute(fp, parameter);

        assertSame(h, service.getStringHolder());

        verifyControls();
    }

    public void testAutowireServicesFailure()
    {
        MockControl fpc = newControl(ServiceImplementationFactoryParameters.class);
        ServiceImplementationFactoryParameters fp = (ServiceImplementationFactoryParameters) fpc
                .getMock();

        MockControl mc = newControl(Module.class);
        Module module = (Module) mc.getMock();

        MockControl ehc = newControl(ErrorHandler.class);
        ErrorHandler eh = (ErrorHandler) ehc.getMock();

        Log log = (Log) newMock(Log.class);

        fp.getLog();
        fpc.setReturnValue(log);

        fp.getServiceId();
        fpc.setReturnValue("foo.bar");

        fp.getInvokingModule();
        fpc.setReturnValue(module);

        module.getClassResolver();
        mc.setReturnValue(new DefaultClassResolver());

        module.getService(StringHolder.class);
        mc.setThrowable(new ApplicationRuntimeException("Simulated failure."));

        module.getErrorHandler();
        mc.setReturnValue(eh);

        eh.error(
                log,
                "Unable to autowire property stringHolder of service foo.bar: Simulated failure.",
                null,
                new ApplicationRuntimeException(""));
        ehc.setMatcher(new AggregateArgumentsMatcher(new ArgumentMatcher[]
        { null, null, null, new TypeMatcher() }));

        replayControls();

        BuilderParameter parameter = new BuilderParameter();

        parameter.setClassName(ServiceAutowireTarget.class.getName());
        parameter.setAutowireServices(true);

        ServiceAutowireTarget service = (ServiceAutowireTarget) execute(fp, parameter);

        assertNull(service.getStringHolder());

        verifyControls();
    }

    public void testAutowireConstructor() throws Exception
    {
        MockControl fpc = newControl(ServiceImplementationFactoryParameters.class);
        ServiceImplementationFactoryParameters fp = (ServiceImplementationFactoryParameters) fpc
                .getMock();

        MockControl mc = newControl(Module.class);
        Module module = (Module) mc.getMock();

        MockControl lc = newControl(Log.class);
        Log log = (Log) lc.getMock();

        fp.getLog();
        fpc.setReturnValue(log);

        fp.getServiceId();
        fpc.setReturnValue("foo");

        fp.getInvokingModule();
        fpc.setReturnValue(module, MockControl.ONE_OR_MORE);

        ClassResolver classResolver = new DefaultClassResolver();

        module.getClassResolver();
        mc.setReturnValue(classResolver);

        module.containsService(Comparable.class);
        mc.setReturnValue(false);

        module.containsService(StringHolder.class);
        mc.setReturnValue(true);

        StringHolder h = new StringHolderImpl();

        module.getService(StringHolder.class);
        mc.setReturnValue(h);

        module.getClassResolver();
        mc.setReturnValue(classResolver);

        replayControls();

        BuilderParameter parameter = new BuilderParameter();

        parameter.setClassName(ConstructorAutowireTarget.class.getName());
        parameter.setAutowireServices(true);
        parameter.addProperty(new BuilderClassResolverFacet());

        ConstructorAutowireTarget service = (ConstructorAutowireTarget) execute(fp, parameter);

        assertSame(h, service.getStringHolder());
        assertSame(classResolver, service.getClassResolver());

        verifyControls();
    }

    public void testAutowireConstructorFailure() throws Exception
    {
        MockControl fpc = newControl(ServiceImplementationFactoryParameters.class);
        ServiceImplementationFactoryParameters fp = (ServiceImplementationFactoryParameters) fpc
                .getMock();

        MockControl mc = newControl(Module.class);
        Module module = (Module) mc.getMock();

        MockControl elc = newControl(ErrorLog.class);
        ErrorLog el = (ErrorLog) elc.getMock();

        MockControl lc = newControl(Log.class);
        Log log = (Log) lc.getMock();

        fp.getLog();
        fpc.setReturnValue(log);

        fp.getServiceId();
        fpc.setReturnValue("foo");

        fp.getInvokingModule();
        fpc.setReturnValue(module, MockControl.ONE_OR_MORE);

        ClassResolver classResolver = new DefaultClassResolver();

        module.getClassResolver();
        mc.setReturnValue(classResolver);

        module.containsService(Comparable.class);
        mc.setReturnValue(false);

        module.containsService(StringHolder.class);
        mc.setReturnValue(false, 2);

        fp.getErrorLog();
        fpc.setReturnValue(el);

        el.error(
                "Error building service foo: Unable to find constructor applicable for autowiring. "
                        + "Use explicit constructor parameters.",
                null,
                new ApplicationRuntimeException(""));
        elc.setMatcher(new AggregateArgumentsMatcher(new ArgumentMatcher[]
        { null, null, new TypeMatcher() }));

        replayControls();

        BuilderParameter parameter = new BuilderParameter();

        parameter.setClassName(ConstructorAutowireTarget.class.getName());
        parameter.setAutowireServices(true);

        assertNull(execute(fp, parameter));

        verifyControls();
    }

    public void testSetObject() throws Exception
    {
        Registry r = buildFrameworkRegistry("SetObject.xml");

        SetObjectFixture f = (SetObjectFixture) r.getService(SetObjectFixture.class);

        assertNotNull(f.getClassFactory1());
        assertSame(f.getClassFactory1(), f.getClassFactory2());
    }

    public void testAutowireService() throws Exception
    {
        Registry r = buildFrameworkRegistry("AutowireService.xml");

        SetObjectFixture f = (SetObjectFixture) r.getService(SetObjectFixture.class);

        assertNotNull(f.getClassFactory1());
        assertSame(f.getClassFactory1(), f.getClassFactory2());
    }

}